"""
    `confluo` - Scalable distributed microservice framework using AMQP/RabbitMQ

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""

__VERSION__ = "0.0.1"

from .core import Service


__all__ = ["Service"]
